import logging
import random
from time import perf_counter

import networkx as nx
from networkx.algorithms.shortest_paths.weighted import _weight_function

from evrouting.graph_tools import DISTANCE_KEY

from evaluation.lib.algorithm import _dijkstra_multisource
from evaluation.lib.export import write_head, write_row
from evaluation.lib import queries

logger = logging.getLogger(__name__)

file_names = {
    queries.classic_query: 'classic.csv',
    queries.astar_query: 'astar.csv',
    queries.bidirectional_query: 'bidirectional.csv',
    queries.charge_query: 'charge.csv',
    queries.gasstation_query: 'gasstation.csv',
    queries.init_gasstation_queries: 'init.csv',
    queries.insert_nodes_into_state_graph: 'insert.csv'
}

conf_algorithms = {
    'classic': queries.classic_query,
    'astar': queries.astar_query,
    'bidirectional': queries.bidirectional_query,
    'charge': queries.charge_query,
    'gasstation': queries.gasstation_query
}


def _insert_charging_stations(graph, charging_stations, number=None):
    start = perf_counter()
    graph.insert_charging_stations(charging_stations, number)
    runtime = perf_counter() - start

    logger.info('Importing {} Charging Stations took {:.2f} s'.format(
        len(graph.charging_stations),
        runtime
    ))


def _init_result_files(result_dir, q=None):
    if q is None:
        q = conf_algorithms.values()

    # Remove existing results and write heads
    for query in q:
        with result_dir.joinpath(file_names[query]).open('w') as f:
            write_head(f, queries.row_classes[query])


def _run_queries(graph, start_nodes, target_nodes, setup, q, result_dir):
    for func in q:
        logger.info(f'Running {len(start_nodes)} times {func.__name__}')

        for i, (s, t) in enumerate(zip(start_nodes, target_nodes)):
            logger.debug(f'{i + 1}/{len(start_nodes)}')
            result_data = func(graph, setup, s, t)
            with result_dir.joinpath(file_names[func]).open('a') as f:
                write_row(f, result_data)

        # Delete cached graphs
        for key in list(queries.CACHE.keys()):
            del queries.CACHE[key]


def _get_target_with_rank(graph, start_node, rank):
    weight = _weight_function(graph, DISTANCE_KEY)
    return _dijkstra_multisource(
        graph,
        [start_node],
        weight=weight,
        rank=rank
    )


def query(graphs, charging_stations, conf, result_dir):
    _init_result_files(result_dir)
    for map_name, G in zip(conf['maps'], graphs):
        for setup in conf['setups']:
            nodes = random.sample(list(G.nodes), k=2 * conf['queries_per_setup'])
            # Random start and target nodes
            start_nodes = nodes[:int(len(nodes) / 2)]
            target_nodes = nodes[int(len(nodes) / 2):]

            # Random adding of charging stations
            _insert_charging_stations(G, charging_stations, setup['charging_stations'])

            # Get algorithms for this setup
            algorithms = setup.get('algorithms', [
                'classic', 'astar', 'bidirectional', 'gasstation', 'charge'
            ])
            q = [conf_algorithms[key] for key in algorithms]

            logger.info(f"Running queries on map {map_name}")
            _run_queries(G, start_nodes, target_nodes, setup, q, result_dir)


def rank(graphs, charging_stations, conf, result_dir):
    ranks = conf['ranks']
    queries_per_rank = conf['queries_per_rank']
    algorithms = ['classic', 'charge']
    q = [conf_algorithms[k] for k in algorithms]

    _init_result_files(result_dir, q)

    for graph in graphs:
        _insert_charging_stations(graph, charging_stations)

        for r in ranks:
            start_nodes = random.sample(list(graph.nodes), queries_per_rank)

            logger.debug(
                f'Getting {queries_per_rank} target nodes with rank {r}.'
            )

            target_nodes = []
            for s in start_nodes:
                try:
                    target_nodes.append(_get_target_with_rank(graph, s, r))
                except nx.NetworkXNoPath:
                    start_nodes = start_nodes[:len(target_nodes)]
                    break

            for setup in conf['setups']:
                _run_queries(graph, start_nodes, target_nodes,
                             setup, q, result_dir)


def init(graphs, charging_stations, conf, result_dir):
    _init_result_files(result_dir, [
        queries.init_gasstation_queries,
        queries.insert_nodes_into_state_graph
    ])

    for map_name, G in zip(conf['maps'], graphs):
        nodes = random.sample(list(G.nodes), k=2 * conf['queries_per_setup'])
        start_nodes = nodes[:int(len(nodes) / 2)]
        target_nodes = nodes[int(len(nodes) / 2):]

        for setup in conf['setups']:
            # Random adding of charging stations
            _insert_charging_stations(G, charging_stations, setup['charging_stations'])

            result_data = queries.init_gasstation_queries(G, setup)
            filename = file_names[queries.init_gasstation_queries]
            with result_dir.joinpath(filename).open('a') as f:
                write_row(f, result_data)

            # random start and target nodes of state graph
            _run_queries(
                G,
                start_nodes,
                target_nodes,
                setup,
                [queries.insert_nodes_into_state_graph],
                result_dir
            )

            # Delete cached graphs
            for key in list(queries.CACHE.keys()):
                del queries.CACHE[key]
