import logging
import gc
from collections import namedtuple
from dataclasses import dataclass

from time import perf_counter

import networkx as nx

from evrouting import gasstation, charge
from evrouting.graph_tools import (
    consumption_function_distance_factory,
    consumption_function_time_factory,
    DISTANCE_KEY
)
from evrouting.osm.profiles import car
from evrouting.osm.routing import GasstationAccessFunctions, a_start_heuristic

from evaluation.lib.algorithm import ranked_dijkstra


@dataclass
class InitQueryRow:
    time_contracted_graph: float
    time_state_graph: float
    nodes_state_graph: int
    edges_state_graph: int
    nodes_contracted_graph: int
    edges_contracted_graph: int
    charging_stations: int


@dataclass
class InsertQueryRow:
    start_node: str
    target_node: str
    charging_stations: int
    time_insert_start: float
    time_insert_target: float


@dataclass
class QueryRow:
    start_node: str
    target_node: str
    query_time: float
    trip_time: float
    nodes: int
    edges: int
    charging_stations: int


@dataclass
class ChargingStopsQuery(QueryRow):
    charging_stops: int
    charging_time: float


@dataclass
class GasstationQueryRow(ChargingStopsQuery):
    time_contracted_graph: float
    time_state_graph: float
    nodes_state_graph: int
    edges_state_graph: int


@dataclass
class ChargeQueryRow(ChargingStopsQuery):
    pass


@dataclass
class ClassicQueryRow(QueryRow):
    dijkstra_rank: int


@dataclass
class AStarQueryRow(QueryRow):
    pass


CACHE = {}


def cached(key):
    """Cache return value."""

    def decorator(func):
        def getter(*args, **kwargs):
            try:
                return CACHE[key]
            except KeyError:
                ret = func(*args, **kwargs)
                CACHE[key] = ret
                return ret

        return getter

    return decorator


def no_gc(func):
    """Run func without garbage collection."""

    def inner(*args, **kwargs):
        gcold = gc.isenabled()
        gc.disable()
        try:
            r = func(*args, **kwargs)
        finally:
            if gcold:
                gc.enable()
        return r

    inner.__name__ = func.__name__

    return inner


_cache_contraction_graph_key = 'contraction'
_cache_state_graph_key = 'state_graph'


@cached(_cache_contraction_graph_key)
def get_contracted_graph(graph, conf, f):
    start = perf_counter()
    contracted_graph = gasstation.routing.contract_graph(
        G=graph,
        charging_stations=graph.charging_stations,
        capacity=conf['capacity'],
        f=f
    )
    contraction_time = perf_counter() - start
    logging.info('Contracted graph with {} stations in {:.2f} s.'.format(
        len(graph.charging_stations),
        contraction_time
    ))
    return contracted_graph, contraction_time


@cached(_cache_state_graph_key)
def get_state_graph(contracted_graph, conf, f):
    start = perf_counter()
    state_graph = gasstation.routing.state_graph(
        contracted_graph,
        conf['capacity'],
        f
    )
    state_graph_time = perf_counter() - start
    logging.info('Created State Graph in {:.2f} s.'.format(
        state_graph_time
    ))
    return state_graph, state_graph_time


@no_gc
def init_gasstation_queries(graph, conf):
    f = GasstationAccessFunctions(conf['consumption']['consumption_coefficient'])
    contracted_graph, contraction_time = get_contracted_graph(graph, conf, f)
    state_graph, state_graph_time = get_state_graph(contracted_graph, conf, f)

    return InitQueryRow(
        charging_stations=len(graph.charging_stations),
        time_contracted_graph=contraction_time,
        time_state_graph=state_graph_time,
        nodes_state_graph=len(state_graph.nodes),
        edges_state_graph=len(state_graph.edges),
        nodes_contracted_graph=len(contracted_graph.nodes),
        edges_contracted_graph=len(contracted_graph.edges)
    )


@no_gc
def insert_nodes_into_state_graph(graph, conf, s, t):
    f = GasstationAccessFunctions(conf['consumption']['consumption_coefficient'])
    contracted_graph, contraction_time = get_contracted_graph(graph, conf, f)
    state_graph, state_graph_time = get_state_graph(contracted_graph, conf, f)

    start = perf_counter()
    gasstation.routing.insert_start_node(
        s,
        graph,
        contracted_graph,
        graph.charging_stations,
        state_graph,
        conf['capacity'],
        conf['mu_s'],
        f
    )
    time_insertion = perf_counter() - start

    start = perf_counter()
    gasstation.routing.insert_final_node(
        t,
        graph,
        graph.charging_stations,
        state_graph,
        conf['capacity'],
        conf['mu_t'],
        f
    )
    time_target = perf_counter() - start

    return InsertQueryRow(
        start_node=s[0],
        target_node=t[0],
        charging_stations=len(graph.charging_stations),
        time_insert_start=time_insertion,
        time_insert_target=time_target
    )


@no_gc
def gasstation_query(graph, conf, s, t):
    f = GasstationAccessFunctions(conf['consumption']['consumption_coefficient'])
    contracted_graph, contraction_time = get_contracted_graph(graph, conf, f)
    state_graph, state_graph_time = get_state_graph(contracted_graph, conf, f)

    start = perf_counter()
    result = gasstation.routing.shortest_path(
        G=graph,
        charging_stations=graph.charging_stations,
        s=s,
        t=t,
        initial_soc=conf['mu_s'],
        final_soc=conf['mu_t'],
        capacity=conf['capacity'],
        f=f,
        extended_graph=state_graph,
        contracted_graph=contracted_graph
    )
    query_time = perf_counter() - start

    charge_stops = [t for n, t in result.charge_path if t > 0]
    return GasstationQueryRow(
        start_node=s,
        target_node=t,
        query_time=query_time,
        trip_time=result.trip_time,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations),
        time_contracted_graph=contraction_time,
        time_state_graph=state_graph_time,
        charging_stops=len(charge_stops),
        charging_time=sum(charge_stops),
        nodes_state_graph=len(state_graph.nodes),
        edges_state_graph=len(state_graph.edges)
    )


@no_gc
def charge_query(graph, conf, s, t):
    start = perf_counter()
    if conf['consumption'].get('type', 'distance') == 'time':
        c = consumption_function_time_factory(conf['consumption']['consumption_coefficient'])
    else:
        c = consumption_function_distance_factory(conf['consumption']['consumption_coefficient'])
    result = charge.shortest_path(
        G=graph,
        charging_stations=graph.charging_stations,
        s=s,
        t=t,
        initial_soc=conf['mu_s'],
        final_soc=conf['mu_t'],
        capacity=conf['capacity'],
        c=c
    )
    runtime = perf_counter() - start
    charge_stops = [t for n, t in result.charge_path if t > 0]
    return ChargeQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result.trip_time,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations),
        charging_stops=len(charge_stops),
        charging_time=sum(charge_stops)
    )


@no_gc
def classic_query(graph, conf, s, t):
    start = perf_counter()
    try:
        result, rank = ranked_dijkstra(graph, s, t, weight=DISTANCE_KEY)
    except nx.NetworkXNoPath:
        result, rank = None, None
    runtime = perf_counter() - start

    return ClassicQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations),
        dijkstra_rank=rank
    )


@no_gc
def bidirectional_query(graph, conf, s, t):
    start = perf_counter()
    try:
        result, _ = nx.algorithms.bidirectional_dijkstra(
            graph, s, t, weight=DISTANCE_KEY)
    except nx.NetworkXNoPath:
        result = None
    runtime = perf_counter() - start

    return QueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations),
    )


@no_gc
def astar_query(graph, conf, s, t):
    start = perf_counter()
    try:
        result = nx.astar_path_length(
            graph,
            s, t,
            weight=DISTANCE_KEY,
            heuristic=a_start_heuristic(graph, car)
        )
    except nx.NetworkXNoPath:
        result = None
    runtime = perf_counter() - start

    return AStarQueryRow(
        start_node=s,
        target_node=t,
        query_time=runtime,
        trip_time=result,
        nodes=len(graph.nodes),
        edges=len(graph.edges),
        charging_stations=len(graph.charging_stations)
    )


row_classes = {
    classic_query: ClassicQueryRow,
    astar_query: AStarQueryRow,
    bidirectional_query: QueryRow,
    gasstation_query: GasstationQueryRow,
    charge_query: ChargeQueryRow,
    init_gasstation_queries: InitQueryRow,
    insert_nodes_into_state_graph: InsertQueryRow
}
